import { Api } from '@openstapps/gitlab-api';
import { Issue, IssueState, Project, User } from '@openstapps/gitlab-api/lib/types';
/**
 * A structure for template compilation
 */
export interface StructureForTemplate {
    issuesByAssignee: AssigneeWithIssues[];
    meetingDay: string;
}
/**
 * A map
 */
export interface IssuesGroupedByAssigneeId {
    [assigneeId: number]: AssigneeWithIssues;
}
/**
 * An assignee with assigned issues
 */
export interface AssigneeWithIssues {
    assignee: User;
    issueCounts: {
        closed: number;
        opened: number;
    };
    issues: IssueWithMeta[];
    quota: number;
}
/**
 * Issue with meta information
 */
export interface IssueWithMeta extends Issue {
    $branchExists: boolean;
    $closed: boolean;
    $labels: Array<{
        bold: boolean;
        label: string;
    }>;
    $mergeRequestUrl: string;
    $project: string;
    $weeksOpen: number;
}
/**
 * Merge request data
 */
export interface MergeRequestsForProjects {
    [projectId: string]: Array<{
        issue_iid: number;
        web_url: string;
    }>;
}
/**
 * Check if issue state is opened or closed
 *
 * @param state State to check
 */
export declare function issueStateIsOpenedOrClosed(state: IssueState): state is IssueState.OPENED | IssueState.CLOSED;
/**
 * Get merge request URLs from given data
 *
 * @param projectMergeRequests Merge requests data (object containing array of objects)
 * @param projectId Project ID to get data about merge requests for
 * @param issueIid Issue IID in certain project (relative ID, and not issue's GitLab API ID)
 */
export declare function getMergeRequestUrls(projectMergeRequests: MergeRequestsForProjects, projectId: number, issueIid: number): string[];
/**
 * Get issues from all groups with a specific milestone
 *
 * @param api GitLab API to make requests with
 * @param label Label to filter by
 * @param groups List of groups to get issues for
 */
export declare function getIssues(api: Api, label: string, groups: number[]): Promise<Issue[]>;
/**
 * Get IDs of issues with branches for projects
 *
 * @param api GitLab API To make requests with
 * @param projects List of projects
 */
export declare function getIssueBranches(api: Api, projects: Project[]): Promise<{
    [k: string]: number[];
}>;
/**
 * Get issues grouped by assignees
 *
 * @param api GitLab API to make requests with
 * @param label Label to generate report for
 */
export declare function getIssuesGroupedByAssignees(api: Api, label: string): Promise<AssigneeWithIssues[]>;
/**
 * Get next meeting day
 */
export declare function getNextMeetingDay(): string;
/**
 * Get a list of merge requests for projects
 *
 * @param api GitLab API to make requests with
 * @param projects List of projects
 */
export declare function getMergeRequests(api: Api, projects: Project[]): Promise<MergeRequestsForProjects>;
/**
 * Generate a report
 *
 * @param api GitLab API to make requests with
 * @param label Label to generate report for
 * @param template Template to generate report with
 */
export declare function generateReport(api: Api, label: string, template: string): Promise<string>;
/**
 * Generate a markdown report
 *
 * @param api GitLab API to make requests with
 * @param label Label to generate report for
 */
export declare function report(api: Api, label: string): Promise<void>;
