/*
 * Copyright (C) 2018, 2019 StApps
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */
import {Api} from '@openstapps/gitlab-api';
import {Issue, IssueState, MembershipScope, MergeRequestState, Project, User} from '@openstapps/gitlab-api/lib/types';
import {asyncPool} from 'async-pool-native/dist/async-pool';
import * as moment from 'moment';
import {render} from 'mustache';
import {join, resolve} from 'path';
import {cwd} from 'process';
import {flatten2dArray, getProjects, logger, readFilePromisified, writeFilePromisified} from '../common';
import {BOLD_LABELS, GROUPS, LABEL_WEIGHTS} from '../configuration';

/**
 * A structure for template compilation
 */
export interface StructureForTemplate {
  issuesByAssignee: AssigneeWithIssues[];
  meetingDay: string;
}

/**
 * A map
 */
export interface IssuesGroupedByAssigneeId {
  [assigneeId: number]: AssigneeWithIssues;
}

/**
 * An assignee with assigned issues
 */
export interface AssigneeWithIssues {
  assignee: User;
  issueCounts: {
    closed: number;
    opened: number;
  };
  issues: IssueWithMeta[];
  quota: number;
}

/**
 * Issue with meta information
 */
export interface IssueWithMeta extends Issue {
  $branchExists: boolean;
  $closed: boolean;
  $labels: Array<{
    bold: boolean;
    label: string;
  }>;
  $mergeRequestUrl: string;
  $project: string;
  $weeksOpen: number;
}

/**
 * Merge request data
 */
export interface MergeRequestsForProjects {
  [projectId: string]: Array<{
    issue_iid: number;
    web_url: string;
  }>;
}

/**
 * Check if issue state is opened or closed
 *
 * @param state State to check
 */
export function issueStateIsOpenedOrClosed(state: IssueState): state is IssueState.OPENED | IssueState.CLOSED {
  return ['opened', 'closed'].indexOf(state) >= 0;
}

/**
 * Get merge request URLs from given data
 *
 * @param projectMergeRequests Merge requests data (object containing array of objects)
 * @param projectId Project ID to get data about merge requests for
 * @param issueIid Issue IID in certain project (relative ID, and not issue's GitLab API ID)
 */
export function getMergeRequestUrls(projectMergeRequests: MergeRequestsForProjects,
                                    projectId: number,
                                    issueIid: number): string[] {
  if (typeof projectMergeRequests[projectId] === 'undefined' || projectMergeRequests[projectId].length === 0) {
    return [];
  }

  return projectMergeRequests[projectId].filter((obj) => {
    return obj.issue_iid === issueIid;
  }).map((obj) => {
    return obj.web_url;
  });
}

/**
 * Get issues from all groups with a specific milestone
 *
 * @param api GitLab API to make requests with
 * @param label Label to filter by
 * @param groups List of groups to get issues for
 */
export async function getIssues(api: Api, label: string, groups: number[]): Promise<Issue[]> {
  const issueResults = await asyncPool(2, groups, (groupId) => {
    return api.getIssues({
      groupId: groupId,
    });
  });

  const issues = flatten2dArray(issueResults).filter((issue) => {
    return issue.labels.indexOf(label) >= 0;
  });

  logger.log('Fetched ' + issues.length + ' issue(s).');

  return issues;
}

/**
 * Get IDs of issues with branches for projects
 *
 * @param api GitLab API To make requests with
 * @param projects List of projects
 */
export async function getIssueBranches(
  api: Api,
  projects: Project[]): Promise<{ [k: string]: number[] }> {
  const projectBranches: { [k: string]: number[] } = {};

  await asyncPool(2, projects, async (project) => {
    const branches = await api.getBranchesForProject(project.id);

    // extract issue number from branch
    projectBranches[project.id] = branches.map((branch) => {
      return branch.name.split('-')[0];
    }).filter((branchNameStart) => {
      return branchNameStart.match(/^[0-9]+$/);
    }).map((branchNameStart) => {
      return parseInt(branchNameStart, 10);
    });
  });

  return projectBranches;
}

/**
 * Get issues grouped by assignees
 *
 * @param api GitLab API to make requests with
 * @param label Label to generate report for
 */
export async function getIssuesGroupedByAssignees(api: Api, label: string): Promise<AssigneeWithIssues[]> {
  const issuesByAssignee: IssuesGroupedByAssigneeId = {};

  const groups = flatten2dArray(await asyncPool(2, GROUPS, async (groupId) => {
    return (await api.getSubGroupsForGroup(groupId)).map((group) => {
      return group.id;
    });
  }));
  groups.push.apply(groups, GROUPS);

  const [issues, projects] = await Promise.all([
    getIssues(api, label, groups),
    getProjects(api, groups),
  ]);

  const issueBranches = await getIssueBranches(api, projects);
  const mergeRequests = await getMergeRequests(api, projects);

  issues.forEach((issue) => {
    if (issue.assignee === null) {
      logger.warn('Issue without assignee!', issue.web_url);
      return;
    }

    if (typeof issuesByAssignee[issue.assignee.id] === 'undefined') {
      issuesByAssignee[issue.assignee.id] = {
        assignee: issue.assignee,
        issueCounts: {
          closed: 0,
          opened: 0,
        },
        issues: [],
        quota: 0,
      };
    }

    if (issue.state === IssueState.REOPENED) {
      issue.state = IssueState.OPENED;
    }

    const issueWithMeta: IssueWithMeta = {
      ...issue,
      ...{
        $branchExists: typeof issueBranches[issue.project_id] !== 'undefined'
          && issueBranches[issue.project_id].indexOf(issue.iid) >= 0,
        $closed: issue.state === IssueState.CLOSED,
        $issue: issue,
        $labels: issue.labels.map((issueLabel: string) => {
          return {
            bold: BOLD_LABELS.includes(issueLabel),
            label: issueLabel,
          };
        }),
        $mergeRequestUrl: getMergeRequestUrls(mergeRequests, issue.project_id, issue.iid)[0],
        $project: issue.web_url.replace('https://gitlab.com/', '').split('/issues/')[0],
        $weeksOpen: moment().diff(moment(issue.created_at), 'weeks'),
      },
    };

    if (issueStateIsOpenedOrClosed(issue.state)) {
      issuesByAssignee[issue.assignee.id].issueCounts[issue.state]++;
      issuesByAssignee[issue.assignee.id].issues.push(issueWithMeta);
    }
  });

  // calculate quota
  Object.keys(issuesByAssignee).forEach((_assigneeId) => {
    const assigneeId = parseInt(_assigneeId, 10);

    issuesByAssignee[assigneeId].quota = Math.floor(
      issuesByAssignee[assigneeId].issueCounts.closed
      / (issuesByAssignee[assigneeId].issueCounts.opened
      + issuesByAssignee[assigneeId].issueCounts.closed) * 100,
    );
  });

  // sort issues by weight of labels and status
  Object.keys(issuesByAssignee).forEach((_assigneeId) => {
    const assigneeId = parseInt(_assigneeId, 10);

    issuesByAssignee[assigneeId].issues.sort((a, b) => {
      let weightA = 0;
      let weightB = 0;

      Object.keys(LABEL_WEIGHTS).forEach((issueLabel) => {
        if (a.labels.indexOf(issueLabel) >= 0) {
          weightA += LABEL_WEIGHTS[issueLabel];
        }

        if (b.labels.indexOf(issueLabel) >= 0) {
          weightB += LABEL_WEIGHTS[issueLabel];
        }
      });

      if (a.state === IssueState.CLOSED) {
        weightA -= 10;
      }

      if (b.state === IssueState.CLOSED) {
        weightB -= 10;
      }

      return weightB - weightA;
    });
  });

  return Object.values(issuesByAssignee);
}

/**
 * Get next meeting day
 */
export function getNextMeetingDay() {
  // get "now"
  const now = moment();

  // get first wednesday of month
  const meetingDayMoment = moment().startOf('month').hour(10).isoWeekday(3);

  while (meetingDayMoment.isBefore(now)) {
    // add one week until meeting day is after now
    meetingDayMoment.add(1, 'weeks');
  }

  const meetingDay = meetingDayMoment.format('YYYY-MM-DD');

  // log found meeting day
  logger.info('Generating report for ' + meetingDay + ' of ' + GROUPS.length + ' group(s)...');

  return meetingDay;
}

/**
 * Get a list of merge requests for projects
 *
 * @param api GitLab API to make requests with
 * @param projects List of projects
 */
export async function getMergeRequests(api: Api,
                                       projects: Project[]): Promise<MergeRequestsForProjects> {
  const projectMergeRequests: MergeRequestsForProjects = {};

  // iterate over projects
  await asyncPool(2, projects, async (project) => {
    // check if project can have merge requests
    if (!project.merge_requests_enabled) {
      return;
    }

    // get all merge requests for project
    const mergeRequests = await api.getMergeRequests(MembershipScope.PROJECTS, project.id, MergeRequestState.OPENED);

    // extract issue number from merge request
    projectMergeRequests[project.id] = mergeRequests.map((mergeRequest) => {
      // keep information about web url too
      return {issue_iid: mergeRequest.source_branch.split('-')[0], web_url: mergeRequest.web_url};
    }).filter((branchNameStartAndUrl) => {
      return branchNameStartAndUrl.issue_iid.match(/^[0-9]+$/);
    }).map((branchNameStartAndUrl) => {
      return {
        issue_iid: parseInt(branchNameStartAndUrl.issue_iid, 10),
        web_url: branchNameStartAndUrl.web_url,
      };
    });
  });

  return projectMergeRequests;
}

/**
 * Generate a report
 *
 * @param api GitLab API to make requests with
 * @param label Label to generate report for
 * @param template Template to generate report with
 */
export async function generateReport(api: Api, label: string, template: string): Promise<string> {
  const issuesGroupedByAssignee = await getIssuesGroupedByAssignees(api, label);

  const structureForTemplate: StructureForTemplate = {
    issuesByAssignee: issuesGroupedByAssignee,
    meetingDay: getNextMeetingDay(),
  };

  return render(
    template,
    structureForTemplate,
  );
}

/**
 * Generate a markdown report
 *
 * @param api GitLab API to make requests with
 * @param label Label to generate report for
 */
export async function report(api: Api, label: string) {
  const meetingDay = getNextMeetingDay();

  const markdown = await generateReport(
    api,
    label,
    (await readFilePromisified(resolve(__dirname, '..', '..', 'templates', 'report.md.mustache'))).toString(),
  );

  let filename = join(cwd(), 'reports', meetingDay + '.md');

  if (label !== 'meeting') {
    filename = join(cwd(), 'reports', `${label}.md`);
  }

  await writeFilePromisified(filename, markdown);

  logger.ok(`Wrote file '${filename}'.`);
}
