/*
 * Copyright (C) 2018 StApps
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */
import {Api} from '@openstapps/gitlab-api';
import {
  AccessLevel,
  IssueState,
  MembershipScope,
  Milestone,
  Project,
  Scope,
} from '@openstapps/gitlab-api/lib/types';
import {asyncPool} from 'async-pool-native/dist/async-pool';
import {flatten2dArray, getProjects, logger} from '../common';
import {GROUPS, NEEDED_LABELS, NEEDED_MILESTONES, NOTE_PREFIX, PROTECTED_BRANCHES, SCHOOLS} from '../configuration';

/**
 * Tidy issues without milestone
 *
 * This will set the milestone of issues without milestone to 'Meeting'.
 *
 * @param api GitLab API instance to use for the requests
 */
export async function tidyIssuesWithoutMilestone(api: Api): Promise<void> {
  // fetch issues without milestone from all groups
  const issueResults = await asyncPool(3, GROUPS, (groupId) => {
    return api.getIssues({
      groupId: groupId,
      milestone: 'No Milestone',
      state: IssueState.OPENED,
    });
  });

  // flatten structure, e.g. put all issues in one array
  const issuesWithoutMilestone = flatten2dArray(issueResults);

  logger.info('Found `' + issuesWithoutMilestone.length + '` issue(s) without milestone.');

  const milestoneCache: { [s: number]: Milestone[] } = {};

  await asyncPool(5, issuesWithoutMilestone, async (issue) => {
    if (typeof milestoneCache[issue.project_id] === 'undefined') {
      milestoneCache[issue.project_id] = await api.getMilestonesForProject(issue.project_id);
    }

    let milestoneId = null;

    milestoneCache[issue.project_id].forEach((milestone) => {
      if (milestone.title === 'Meeting') {
        milestoneId = milestone.id;
      }
    });

    if (milestoneId === null) {
      logger.warn('Milestone `Meeting` was not available for issue ' + issue.title + ' (' + issue.web_url + ').');
      return;
    }

    await api.setMilestoneForIssue(issue, milestoneId);

    logger.log('Milestone was set to `Meeting` for issue ' + issue.title + ' (' + issue.web_url + ').');

    await api.createNote(
      issue.project_id,
      Scope.ISSUES,
      issue.iid,
      `${NOTE_PREFIX} Milestone was set automatically to \`Meeting\`.`,
    );
  });

  logger.ok('Tidied issues without milestones.');
}

/**
 * Tidy open issues without meeting label
 *
 * This adds the label 'meeting' to all open issues that do not have this label.
 *
 * @param api GitLab API instante to use for requests
 */
export async function tidyOpenIssuesWithoutMeetingLabel(api: Api): Promise<void> {
  // fetch all open issues
  const issueResults = await asyncPool(3, GROUPS, (groupId) => {
    return api.getIssues({
      groupId: groupId,
      state: IssueState.OPENED,
    });
  });

  // flatten structure, e.g. put all issues in one array
  const openIssues = flatten2dArray(issueResults);

  logger.info(`Found ${openIssues.length} open issue(s).`);

  // filter issues without meeting label
  const openIssuesWithoutMeetingLabel = openIssues.filter((openIssue) => {
    return openIssue.labels.indexOf('meeting') === -1;
  });

  logger.info(`Filtered ${openIssuesWithoutMeetingLabel.length} open issue(s) without label 'meeting'.`);

  await asyncPool(5, openIssuesWithoutMeetingLabel, async (issue) => {
    if (issue.milestone !== null && issue.milestone.title === 'Backlog') {
      logger.info(`Skipping issue "${issue.title}" because it is in backlog.`);
      return;
    }

    return api.createNote(
      issue.project_id,
      Scope.ISSUES,
      issue.iid,
      `${NOTE_PREFIX} Automatically adding label 'meeting'\n\n/label ~meeting`);
  });

  logger.ok(`Tidied open issues without label 'meeting'.`);
}

/**
 * Tidy labels in a list of projects
 *
 * @param api GitLab API instance to use for the requests
 * @param projects List of projects to tidy labels on
 */
export async function tidyLabels(api: Api, projects: Project[]): Promise<void> {
  await asyncPool(5, projects, async (project) => {
    const labels = await api.getLabels(project.id);

    const neededLabels = NEEDED_LABELS.slice(0);
    // const extraneousLabels: Label[] = [];

    labels.forEach((label) => {
      // let needed = false;

      neededLabels.forEach((neededLabel, neededLabelIdx) => {
        if (neededLabel.name.toLowerCase() === label.name.toLowerCase()) {
          neededLabels.splice(neededLabelIdx, 1);
          // needed = true;
        }
      });

      /* if (!needed) {
        extraneousLabels.push(label);
      } */
    });

    await asyncPool(2, neededLabels, async (neededLabel) => {
      await api.createLabel(project.id, neededLabel.name, neededLabel.description, neededLabel.color);

      logger.log('Created label `' + neededLabel.name + '` in ' + project.name_with_namespace + '.');
    });

    // await asyncPool(2, extraneousLabels, async (extraneousLabel) => {
    //   await api.deleteLabel(project.id, extraneousLabel.name);
    //
    //   logger.log('Deleted label `' + extraneousLabel.name + '` from ' + project.name_with_namespace + '.');
    // });
  });

  logger.ok('Tidied labels.');
}

/**
 * Tidy milestones in a list of projects
 *
 * @param api GitLab API instance to use for the requests
 * @param projects List of projects to tidy milestones on
 */
export async function tidyMilestones(api: Api, projects: Project[]): Promise<void> {
  await asyncPool(5, projects, async (project) => {
    const milestones = await api.getMilestonesForProject(project.id);
    const missingMilestones = NEEDED_MILESTONES.slice(0);

    milestones.forEach((milestone) => {
      const idx = missingMilestones.indexOf(milestone.title);

      if (idx >= 0) {
        missingMilestones.splice(idx, 1);
      }
    });

    if (missingMilestones.length > 0) {
      await asyncPool(2, missingMilestones, async (milestone) => {
        await api.createMilestone(project.id, milestone);
        logger.log('Created milestone ' + milestone + ' for project ' + project.name_with_namespace + '.');
      });
    }
  });

  logger.ok('Tidied milestones.');
}

/**
 * Tidy protected branches in a list of projects
 *
 * @param api GitLab API instance to use for the requests
 * @param projects List of projects to tidy milestones on
 */
export async function tidyProtectedBranches(api: Api, projects: Project[]): Promise<void> {
  await asyncPool(2, projects, async (project) => {
    const branches = await api.getBranchesForProject(project.id);

    const protectableBranches = branches.filter((branch) => {
      return PROTECTED_BRANCHES.indexOf(branch.name) >= 0;
    });

    const unprotectedBranches = protectableBranches.filter((branch) => {
      return !branch.protected;
    });

    await asyncPool(2, unprotectedBranches, async (branch) => {
      await api.protectBranch(project.id, branch.name);

      logger.log('Added protected branch `' + branch.name + '` in project `' + project.name_with_namespace + '`...');
    });
  });

  logger.ok('Tidied protected branches.');
}

/**
 * Tidy "sub" group members
 *
 * @param api GitLab API instance to use for the requests
 */
export async function tidySubGroupMembers(api: Api): Promise<void> {
  const stappsMembers = await api.getMembers(MembershipScope.GROUPS, GROUPS[0]);
  const stappsMemberIds = stappsMembers.map((member) => member.id);

  const groupIdsToSchool: any = {};

  Object.keys(SCHOOLS).map((school) => {
    groupIdsToSchool[SCHOOLS[school]] = school;
  });

  await asyncPool(2, GROUPS.slice(1), async (groupId) => {
    const members = await api.getMembers(MembershipScope.GROUPS, groupId);
    const memberIds = members.map((member) => member.id);

    await asyncPool(2, stappsMembers, async (stappsMember) => {
      if (memberIds.indexOf(stappsMember.id) === -1) {
        await api.addMember(MembershipScope.GROUPS, groupId, stappsMember.id, AccessLevel.Developer);

        logger.log('Added ' + stappsMember.name + ' to group ' + groupIdsToSchool[groupId] + '.');
      }
    });

    await asyncPool(2, members, async (member) => {
      if (stappsMemberIds.indexOf(member.id) === -1) {
        await api.deleteMember(MembershipScope.GROUPS, groupId, member.id);

        logger.log('Deleted member ' + member.name + ' from group ' + groupIdsToSchool[groupId] + '.');
      }
    });
  });

  logger.ok('Tidied "sub" group members.');
}

/**
 * Tidy issues without assignee
 *
 * Set assignee to author if no assignee is set.
 *
 * @param api
 */
export async function tidyIssuesWithoutAssignee(api: Api): Promise<void> {
  // fetch issues without milestone from all groups
  const issueResults = await asyncPool(3, GROUPS, (groupId) => {
    return api.getIssues({
      groupId: groupId,
      state: IssueState.OPENED,
    });
  });

  // flatten structure, e.g. put all issues in one array
  const issues = flatten2dArray(issueResults);

  const issuesWithoutAssignee = issues.filter((issue) => {
    return issue.assignee === null;
  });

  logger.info('Found `' + issuesWithoutAssignee.length + '` issue(s) without assignee.');

  await asyncPool(3, issuesWithoutAssignee, async (issue) => {
    await api.setAssigneeForIssue(issue, issue.author.id);

    logger.log('Set assignee for `' + issue.title + '` to ' + issue.author.name + '.');

    await api.createNote(
      issue.project_id,
      Scope.ISSUES,
      issue.iid,
      `${NOTE_PREFIX} Assignee was set automatically to author.`,
    );
  });

  logger.ok('Tidied issues without assignee.');
}

/**
 * Tidy
 *
 * @param api GitLab API instance to use for the requests
 */
export async function tidy(api: Api) {
  const projects = await getProjects(api, GROUPS);

  await Promise.all([
    tidyLabels(api, projects),
    tidyMilestones(api, projects),
    tidyProtectedBranches(api, projects),
    tidySubGroupMembers(api),
  ]);

  await tidyOpenIssuesWithoutMeetingLabel(api);
  await tidyIssuesWithoutAssignee(api);
}
